local MINOR_VERSION = tonumber(("$Revision: 38684 $"):match("%d+"))

Omen =  AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceConsole-2.0", "AceDB-2.0", "FuBarPlugin-2.0")
Omen:RegisterDB("OmenDB", "OmenDBPC")
Omen:RegisterDefaults('profile', {
	ShowOmen = true,
	AnimateBars = true,
	StretchBarTextures = false,
	BarTexture = "Glaze",
	TPSUpdateFrequency = 0.05,
	TPSPrecision = "%2.1f",
	TPSWindow = 10,
	GrowUp = false,
	MaxBars = 10,
	BarHeight = 15,
	ShowTPS = true,
	BackdropColor = {r = 0, g = 0, b = 0, a = 0.5},
	BorderColor = {r = 0, g = 0, b = 0, a = 0.5},
	PlayerBarColor = {r = 1, g = 0, b = 0, a = 1},
	UseCustomBarColors = false
})

local L = AceLibrary("AceLocale-2.2"):new("Omen")

local Threat = AceLibrary("Threat-1.0")
local SML = AceLibrary("SharedMedia-1.0")

Omen.version = "2.1r" .. MINOR_VERSION
Omen.revision = MINOR_VERSION

local new, newHash, newSet, del
do
	local list = setmetatable({}, {__mode='k'})
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return {...}
		end
	end
	function newHash(...)
		local t = next(list)
		if t then
			list[t] = nil
		else
			t = {}
		end	
		for i = 1, select('#', ...), 2 do
			t[select(i, ...)] = select(i+1, ...)
		end
		return t
	end
	function newSet(...)
		local t = next(list)
		if t then
			list[t] = nil
		else
			t = {}
		end	
		for i = 1, select('#', ...) do
			t[select(i, ...)] = true
		end
		return t
	end
	function del(t)
		setmetatable(t, nil)
		for k in pairs(t) do
			t[k] = nil
		end
		list[t] = true
		return nil
	end
end

local tpsSigma = {}
local lastThreat = {}

function Omen:OnInitialize()
	self.bars = {}

	self:RegisterEvent("Threat_ThreatUpdated")
	self:RegisterEvent("Threat_ThreatCleared", "Clear")
	self:RegisterEvent("Threat_PartyChanged")
	self:RegisterEvent("Threat_OutOfDateNotice")
	self:RegisterEvent("Threat_Activate")
	self:RegisterEvent("Threat_Deactivate")	
	
	self.hasIcon = true
	self:SetIcon("Interface\\AddOns\\Omen\\icon")
	self.Anchor = CreateFrame("Frame", "OmenAnchor", UIParent)
	self.Anchor:SetWidth(210)
	self.Anchor:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
	self.Anchor:SetBackdrop({
		bgFile = "Interface/Tooltips/UI-Tooltip-Background",
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
		tile = true,
		tileSize = 16,
		edgeSize = 16,
		insets = {left = 4, right = 4, top = 4, bottom = 4}
	})
	self.Anchor:SetClampedToScreen(true)
	self.Anchor.title = self.Anchor:CreateFontString(nil, nil, "GameFontNormal")
	self.Anchor.title:SetPoint("BOTTOM", self.Anchor, "TOP", 0, 2)

	self.Anchor.revision = self.Anchor:CreateFontString(nil, nil, "GameFontNormal")
	self.Anchor.revision:SetPoint("TOPLEFT", self.Anchor, "BOTTOMLEFT", 4, 2)
	local font, size = self.Anchor.revision:GetFont()
	self.Anchor.revision:SetFont(font, 8)
	self.modName = "|cff6299FFOmen"
	self.Anchor.revision:SetText(string.format("|cffFFB400Omen r |cffffffff%s /|cffFFB400 Threat-1.0 r |cffffffff%s", MINOR_VERSION, select(2,Threat:GetLibraryVersion())))

	self.Anchor.outofDate = self.Anchor:CreateFontString(nil, nil, "GameFontNormal")
	self.Anchor.outofDate:SetPoint("TOPLEFT", self.Anchor.revision, "BOTTOMLEFT", 0, -1)
	self.Anchor.outofDate:Hide()
	local font, size = self.Anchor.revision:GetFont()
	self.Anchor.outofDate:SetFont(font, 8)
	self.Anchor.outofDate:SetText("|cffff0000Out of date notice")
	
	self.threatTargetName = ""
	self.Anchor.title:SetText(self.modName)
	local f = self.db.profile.BackdropColor
	self.Anchor:SetBackdropColor(f.r, f.g, f.b, f.a)

	local f = self.db.profile.BorderColor
	self.Anchor:SetBackdropBorderColor(f.r, f.g, f.b, f.a)

	self.barsOffset = 22

	self.Anchor:SetMovable(true)
	self.Anchor:EnableMouse(true)
	self.Anchor:SetScript("OnMouseDown", function() this:StartMoving(); end)
	self.Anchor:SetScript("OnMouseUp", function() this:StopMovingOrSizing(); end)
	self.barList = {}
	self:Toggle(self.db.profile.ShowOmen)
	self:RegisterBucketEvent("Omen_ThreatUpdated", 0.2, "ArrangeBars")
	self:ArrangeBars()
	
	self.Anchor:SetHeight(self.barsOffset + 5)
	
	local t = self.Anchor:CreateFontString(nil, nil, "GameFontNormal")
	t:SetVertexColor(1,1,1,1)
	local font, size = t:GetFont()
	t:SetFont(font, 11)
	t:SetText(L["Name"])
	self.Anchor.nameLabel = t

	local t = self.Anchor:CreateFontString(nil, nil, "GameFontNormal")
	t:SetVertexColor(1,1,1,1)
	local font, size = t:GetFont()
	t:SetFont(font, 11)
	t:SetText(L["TPS"])
	if not self.db.profile.ShowTPS then
		t:Hide()
	end
	self.Anchor.tpsLabel = t

	local t = self.Anchor:CreateFontString(nil, nil, "GameFontNormal")
	t:SetVertexColor(1,1,1,1)
	local font, size = t:GetFont()
	t:SetFont(font, 11)
	t:SetText(L["Threat"])
	self.Anchor.totalLabel = t
	
	self.Anchor.nameLabel:SetPoint("TOPLEFT", self.Anchor, "TOPLEFT", 7, -7)
	self.Anchor.tpsLabel:SetPoint("TOPLEFT", self.Anchor, "TOPLEFT", self.Anchor:GetWidth() * 0.55 + 5, -7)
	self.Anchor.totalLabel:SetPoint("TOPRIGHT", self.Anchor, "TOPRIGHT", -7, -7)
end

local function ConvertToK(n)
	if n > 1000 then
		return n / 1000, "k"
	end
	return n, ""
end

local default_color = {r=0,g=1,b=1,a=1}
function addBar(self, name, unit)
	if not self.bars[name] then
		local _, cls = UnitClass(unit)
		local c = RAID_CLASS_COLORS[cls] or default_color
		self.bars[name] = self:CreateBar()
		if unit == "player" and self.db.profile.UseCustomBarColors then
			local c = self.db.profile.PlayerBarColor
			self.bars[name].tex:SetVertexColor(c.r, c.g, c.b)
		else
			self.bars[name].tex:SetVertexColor(c.r, c.g, c.b)
		end
		self.bars[name].unitid = unit
		self.bars[name].unitname = name
		tinsert(self.barList, self.bars[name])
	end
end

local function resetTPSSigmas(self)
	self:DisableTPSUpdate()
	for k in pairs(tpsSigma) do
		tpsSigma[k] = nil
	end
	for k in pairs(lastThreat) do
		lastThreat[k] = nil
	end
end

function Omen:Threat_Activate()
	self:RegisterEvent("UNIT_TARGET")
	Omen:Toggle(true)
end

function Omen:Threat_Deactivate()
	self:UnregisterEvent("UNIT_TARGET")
	Omen:Toggle(false)
end

function Omen:Toggle(state)
	if state or not self.Anchor:IsVisible() then
		self.Anchor:Show()
		self:ArrangeBars()
	else
		self.Anchor:Hide()
	end

	self.db.profile.ShowOmen = (Omen.Anchor:IsVisible() and true or false) -- DB didn't like nil
end

-- If our party changes around, we want to follow our target by name
-- We use the Threat event rather than the Blizzard event because this way we
-- know that Threat:GetUnitIDByUnitName will be accurate
function Omen:Threat_PartyChanged()
	if self.partyMemberName then
		local partyID = Threat:GetUnitIDByUnitName(self.partyMemberName)
		if partyID then
			self.watchedUnitID = partyID
			self.threatTargetName = UnitName(partyID .. "target")
		else
			-- We couldn't find a corresponding unit for the party member we want to watch
			-- Go ahead and invalidate them
			self.partyMemberName = nil
			self.watchedUnitID = nil
			self.threatTargetName = nil
		end
		self:UpdateThreatLists(self.threatTargetName or Threat.GlobalTarget)
	end
end

function Omen:SetWatchedUnit()
	local targetName = UnitName("target")
	local partyID
	if targetName then
		partyID = Threat:GetUnitIDByUnitName(targetName)
	end
	if partyID then
		self.watchingParty = true
		self.watchedUnitID = partyID
		self.partyMemberName = targetName
		self.threatTargetName = UnitName(partyID .. "target")
	elseif targetName then
		self.watchingParty = false
		self.watchedUnitID = "target"
		self.partyMemberName = nil
		self.threatTargetName = targetName
	else
		self.watchingParty = false
		self.watchedUnitID = nil
		self.partyMemberName = nil
		self.threatTargetName = targetName
	end
	self:UpdateThreatLists(self.threatTargetName or Threat.GlobalTarget)
	self.watchedUnitName = UnitName("target")	
end

function Omen:UpdateThreatLists(target)
	self:Clear()
	self.Anchor.title:SetText(target and string.format("%s - |cffFF4200%s", self.modName, target) or self.modName)
	for k, v in Threat:IterateGroupThreatForTarget(target) do
		if not self.bars[k] then
			local unit = Threat:GetUnitIDByUnitName(k)
			if unit then
				addBar(self, k, unit)
				if unit == "player" then
					self.bars[k].myArrow:Show()
				end				
				local foreign = ""
				if Threat:IsUsingForeignThreatSource(k) then
					foreign = "*"
				end
				self.bars[k].name:SetText(k .. foreign)
			end
		end
		
		local val, isK = ConvertToK(v)
		self.bars[k].threatVal = v
		self.bars[k].threat:SetText(string.format("%2.2f%s", val, isK))
	end
	
	self:TriggerEvent("Omen_ThreatUpdated")
end

function Omen:Clear()
	if not self.bars then return end
	for k, v in pairs(self.bars) do
		v.threatVal = 0
		v:Hide()
	end
	self.Anchor:SetHeight(self.barsOffset + 5)
end

local function func(self, target)
	self.lastWithAggro = target
	self:ArrangeBars()
end
function Omen:UNIT_TARGET(arg1)
	local watchedUnitID = self.watchedUnitID 

	if arg1 == "player" then
		self:SetWatchedUnit()
	elseif arg1 ~= nil and watchedUnitID then
		local watchingParty = self.watchingParty
		local threatTargetName = self.threatTargetName or Threat.GlobalTarget
		
		if watchingParty and UnitIsUnit(arg1, watchedUnitID) then
			self.threatTargetName = UnitName(self.watchedUnitID .. "target")
			self:UpdateThreatLists(threatTargetName)
		elseif not (self.watchingParty and UnitIsUnit(arg1, self.watchedUnitID)) or (self.watchingParty and UnitIsUnit(arg1, self.watchedUnitID .. "target")) then
			local aName = UnitName(arg1 .. "target")
			if aName then
				self:ScheduleEvent("OmenSetAggroTarget", func, 1.5, self, aName)
			end
		end
	end
end

function Omen:Threat_ThreatUpdated(pName, pUnit, target, threatVal)
	self:EnableTPSUpdate()
	self:ScheduleEvent("ResetTPSTimerTables", resetTPSSigmas, 2, self)
	
	if not target then
		self:Print(string.format("Error: Got a nil target, %s, %s, %s", pName, pUnit, threatVal))
		return
	end
	if not self.threatTargetName or (target ~= Threat.GlobalTarget and not Threat:UnitIsUnit(self.threatTargetName, target)) then return end

	if not self.bars[pName] then
		addBar(self, pName, pUnit)
		if pUnit == "player" then
			self.bars[pName].myArrow:Show()
		end
		local foreign = ""
		if Threat:IsUsingForeignThreatSource(pName) then
			foreign = "*"
		end
		self.bars[pName].name:SetText(pName .. foreign)
	end
	if tps then
		self.bars[pName].threatPerSec:SetText(string.format("%2.1f", tps)) 
	end
	local isK = ""
	self.bars[pName].threatVal = tonumber(threatVal)
	threatVal, isK = ConvertToK(threatVal)
	self.bars[pName].threat:SetText(string.format("%2.2f%s", threatVal, isK))
	self:TriggerEvent("Omen_ThreatUpdated")
end

function Omen:Threat_OutOfDateNotice(myVersion, newVersion, newVersionHolder)
	self.Anchor.outofDate:SetText(string.format("|cffff0000New version available! |cffFFB400r |cffffffff%s (%s)", newVersion, newVersionHolder))
	self.Anchor.outofDate:Show()
end

function Omen:CreateAggroGainBar(pName, name, multiplier)
	local aggroToBeat = 0

	if not self.watchedUnitID then return end
	for k, v in pairs(self.barList) do
		v.aggroArrow:Hide()
		if v.unitid and v.unitname == self.lastWithAggro then
			aggroToBeat = v.threatVal
			v.aggroArrow:Show()
			if Threat:UnitInMeleeRange(self.watchedUnitID) then
				multiplier = 1.1
			end
		end
	end
	
	if not self.bars[pName] then
		self.bars[pName] = self:CreateBar()
		local color = (multiplier / 1.3)
		self.bars[pName].tex:SetVertexColor(color, 0, 0)
		self.bars[pName].name:SetText(name)
		tinsert(self.barList,self.bars[pName])
	end
	self.bars[pName].threatVal = aggroToBeat * multiplier
	local threatVal, isK = ConvertToK(self.bars[pName].threatVal)
	self.bars[pName].threat:SetText(string.format("%2.2f%s", threatVal, isK))
end

function Omen:ArrangeBars()
	
	if not self.Anchor:IsVisible() then return end
	local threatTarget = self.watchedUnitID

	if threatTarget then	
		if Threat:UnitInMeleeRange(threatTarget) then
			self:CreateAggroGainBar("_AGGRO_GAIN_", string.format("*%s*", L["Aggro Gain"]), 1.1)
		else
			self:CreateAggroGainBar("_AGGRO_GAIN_", string.format("*%s*", L["Aggro Gain"]), 1.3)
		end
	elseif self.bars["_AGGRO_GAIN_"] then
		self.bars["_AGGRO_GAIN_"].threatVal = 0
	end
	
	table.sort(self.barList, function(a,b)
		return b.threatVal < a.threatVal
	end)
	local lastBar = nil
	local maxThreat = -1
	local barCt = 0
	for k,v in pairs(self.barList) do
		if barCt >= self.db.profile.MaxBars then
			v:Hide()
		else
			if maxThreat < v.threatVal then
				maxThreat = v.threatVal
				if maxThreat == 0 then maxThreat = 1 end
			end
	
			if v.threatVal > 0 then
				v:Show()
				v:ClearAllPoints()
				v:SetPercent(v.threatVal / maxThreat)
				if self.db.profile.GrowUp then
					if lastBar then
						v:SetPoint("BOTTOMLEFT", lastBar, "TOPLEFT", 0, 0)
						v:SetPoint("BOTTOMRIGHT", lastBar, "TOPRIGHT", 0, 0)
					else
						v:SetPoint("BOTTOMLEFT", self.Anchor, "BOTTOMLEFT", 5, self.barsOffset * -1)
						v:SetPoint("BOTTOMRIGHT", self.Anchor, "BOTTOMRIGHT", -5, self.barsOffset * -1)
					end
				else
					if lastBar then
						v:SetPoint("TOPLEFT", lastBar, "BOTTOMLEFT", 0, 0)
						v:SetPoint("TOPRIGHT", lastBar, "BOTTOMRIGHT", 0, 0)
					else
						v:SetPoint("TOPLEFT", self.Anchor, "TOPLEFT", 5, self.barsOffset * -1)
						v:SetPoint("TOPRIGHT", self.Anchor, "TOPRIGHT", -5, self.barsOffset * -1)
					end
				end
				lastBar = v
				barCt = barCt + 1
			else
				v:Hide()
			end
		end
	end
	self.Anchor:SetHeight(self.db.profile.BarHeight * max(0,barCt) + self.barsOffset + 5)
end

local function animateBar()
	this.animatedTime = this.animatedTime + arg1
	local pct = min(1, this.animatedTime / this.totalAnimateTime)
	local delta = (this.destinationWidth - this.startWidth) * pct
	-- Omen:Print(string.format("Setting tex width: %s, %s for %s ", (this.animatedTime / this.totalAnimateTime), this.startWidth + delta, tostring(arg1)))
	this.tex:SetWidth(this.startWidth + delta)
	if not Omen.db.profile.StretchBarTextures then
		local texPct = (this.startWidth + delta) / this.totalWidth
		this.tex:SetTexCoord(0, texPct, 0, 1)
	end
	if pct == 1 then
		this:SetScript("OnUpdate", nil)
	end
end

local function animatePercent(self, pct) 
	local w = self:GetParent():GetWidth() - 10
	local width = w * pct
	if Omen.db.profile.AnimateBars then
		self:SetScript("OnUpdate", animateBar)
		self.totalWidth = w
		self.startWidth = self.tex:GetWidth()
		self.destinationWidth = width
		self.totalAnimateTime = 0.25
		self.animatedTime = 0
	else
		self.tex:SetWidth(width)
		if not Omen.db.profile.StretchBarTextures then
			self.tex:SetTexCoord(0, pct, 0, 1)
		end
	end	
end

function Omen:OnClick()
	self:Toggle()
end

function Omen:CreateBar()
	local f = CreateFrame("Frame", "OmenBar" .. #self.bars, self.Anchor)
	f:SetHeight(self.db.profile.BarHeight)
	
	f.tex = f:CreateTexture()
	f.tex:SetTexture(SML:Fetch("statusbar", Omen.db.profile.BarTexture))
	f.tex:SetPoint("TOPLEFT", f, "TOPLEFT", 0, 0)
	f.tex:SetPoint("BOTTOMLEFT", f, "BOTTOMLEFT", 0, 0)
	f.tex:SetWidth(200)
	
	f.name 		= f:CreateFontString(nil, nil, "GameFontNormal")
	f.name:SetVertexColor(1,1,1,1)

	f.threat 	= f:CreateFontString(nil, nil, "GameFontNormal")
	f.threat:SetVertexColor(1,1,1,1)

	f.threatPerSec 	= f:CreateFontString(nil, nil, "GameFontNormal")
	f.threatPerSec:SetVertexColor(1,1,1,1)
	
	f.myArrow = f:CreateTexture()
	f.myArrow:SetTexture("Interface\\AddOns\\Omen\\arrow")
	f.myArrow:SetPoint("RIGHT", f, "LEFT", -5, 0)
	f.myArrow:SetTexCoord(0, 0.5, 0, 1)
	f.myArrow:SetWidth(6)
	f.myArrow:SetHeight(12)
	f.myArrow:Hide()
	
	f.aggroArrow = f:CreateTexture()
	f.aggroArrow:SetTexture("Interface\\AddOns\\Omen\\arrow")
	f.aggroArrow:SetPoint("LEFT", f, "RIGHT", 5, 0)
	f.aggroArrow:SetTexCoord(0.5, 1, 0, 1)
	f.aggroArrow:SetWidth(6)
	f.aggroArrow:SetHeight(12)	
	f.aggroArrow:Hide()
	
	f.name:SetPoint("LEFT", f, "LEFT", 5, 0)
	local font, size = f.name:GetFont()
	f.name:SetFont(font, 11)

	f.threat:SetPoint("RIGHT", f, "RIGHT", 0, 0)
	local font, size = f.threat:GetFont()
	f.threat:SetFont(font, 10)

	f.threatPerSec:SetPoint("LEFT", f, "LEFT", self.Anchor:GetWidth() * 0.55, 0)
	local font, size = f.threatPerSec:GetFont()
	f.threatPerSec:SetFont(font, 10)
	if not Omen.db.profile.ShowTPS then
		f.threatPerSec:Hide()
	end

	f.SetPercent = animatePercent
	return f
end

------------------------------------------------------------------------------------------------
-- TPS crap
------------------------------------------------------------------------------------------------

function Omen:EnableTPSUpdate()
	if not self:IsEventScheduled("OmenTPSUpdate") then
		self:ScheduleRepeatingEvent("OmenTPSUpdate", "UpdateAllTPS", Omen.db.profile.TPSUpdateFrequency)
	end
end

function Omen:DisableTPSUpdate()
	if self:IsEventScheduled("OmenTPSUpdate") then
		self:CancelScheduledEvent("OmenTPSUpdate")
	end
end

function Omen:UpdateAllTPS()
	local targetName = self.threatTargetName
	if not targetName then return end
	for k, v in pairs(self.bars) do
		if v.threatVal and v.threatVal > 0 and v.unitid then
			self:UpdateTPS(k, targetName)
		end
	end	
end

-- Thanks to Kalman for working out the TPS formula with me.
-- It's just a single-pole IIR filter which decreases the weight of older samples
-- over time. 0 < p1 < 1 - a higher p1 results in faster decay (which means a shorter window)
function Omen:UpdateTPS(player, target)
	local n = target .. "-" .. player
	local sigma = tpsSigma[n]
	local targetThreat = Threat:GetThreat(player, target)
	local p1 = min(0.99, max(0.01, Omen.db.profile.TPSWindow))
	local delta = targetThreat - (lastThreat[n] or 0)
	lastThreat[n] = targetThreat
	if delta < 0 then
		tpsSigma[n] = 0
	else
		tpsSigma[n] = (sigma * p1) + delta
	end
	local tpsDisplay = tpsSigma[n] * (1 - p1) * (1 / self.db.profile.TPSUpdateFrequency)
	self.bars[player].threatPerSec:SetText(string.format("%2.0f", tpsDisplay))
end